import click


class Choice2(click.Choice):
    """ for nice help message """
    def get_metavar(self, param):
        if len(self.choices) == 1:
            return self.choices[0]
        return "[{}]".format("|".join(self.choices))


# classes to create a flag with optional value, see https://stackoverflow.com/a/44144098
# TODO Replace with click 8.0 options? Example:
# @click.option(
#     '--code', is_flag=False, flag_value='_default_',
#     type=click.Choice(['_default_', 'all', 'rejects'])
# )
# Issues:
# 1. _default_ is shown in help text (can be fixed by using a custom Choice class)
# 2. Shared help text for option with and without value.
#    Looks less intuitive than in current implementation. Fix by using custom Command class?

class FlagOption(click.Option):
    """ Mark this option as getting a _opt option """
    is_optflag = True


class OptFlagOption(click.Option):
    """ Fix the help for the _opt suffix """
    def __init__(self, *args, **kwargs):
        # To hide '--optionname_opt' from autocompletoon
        # Assuming that 'hidden', if present, is always in kwargs (in click 8.1 it's the 14th arg)
        kwargs['hidden'] = True
        super().__init__(*args, **kwargs)

    def get_help_record(self, ctx):
        self.hidden = False
        help = super().get_help_record(ctx)
        self.hidden = True
        return (help[0].replace('_opt ', '='),) + help[1:]

    def get_error_hint(self, ctx):
        hint = super().get_error_hint(ctx)
        return hint.replace('_opt', '')


class OptFlagCommand(click.Command):
    """ Command with support for flags with values """
    def parse_args(self, ctx, args):
        """ Translate any flag= to flag_opt= as needed """
        options = [o for o in ctx.command.params
                   if getattr(o, 'is_optflag', None)]
        prefixes = {p for p in sum([o.opts for o in options], [])
                    if p.startswith('--')}
        for i, a in enumerate(args):
            a = a.split('=')
            if a[0] in prefixes and len(a) > 1:
                a[0] += '_opt'
                args[i] = '='.join(a)

        return super().parse_args(ctx, args)


class GroupedGroup(click.Group):
    def add_command(self, command, *args, **kwargs):
        help_group = kwargs.pop('group', None)
        command.help_group = help_group
        return super().add_command(command, *args, **kwargs)

    def format_commands(self, ctx, formatter):
        """see https://stackoverflow.com/a/58770064"""
        commands = []
        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            if not (cmd is None or cmd.hidden):
                commands.append((subcommand, cmd))

        if commands:
            longest = max(len(cmd[0]) for cmd in commands)
            limit = formatter.width - 6 - longest

            groups = {}
            for subcommand, cmd in commands:
                help_str = cmd.get_short_help_str(limit)
                subcommand += ' ' * (longest - len(subcommand))
                groups.setdefault(cmd.help_group, []).append((subcommand, help_str))

            with formatter.section('Commands'):
                for (group_name, _), rows in sorted(groups.items(), key=lambda x: x[0][1]):
                    formatter.write_heading(group_name)
                    with formatter.indentation():
                        formatter.write_dl(rows)
                        formatter.write_paragraph()


class ArgNotRequiredIf(click.Argument):
    # Based on https://stackoverflow.com/a/44349292
    def __init__(self, *args, not_required_if, **kwargs):
        super().__init__(*args, **kwargs)
        self.not_required_if = not_required_if
        assert self.not_required_if, "'not_required_if' parameter required"

    def handle_parse_result(self, ctx, opts, args):
        self_present = opts.get(self.name) is not None
        other_present = opts.get(self.not_required_if) is not None

        if other_present:
            if self_present:
                raise click.UsageError(
                    f"Illegal usage: {self.make_metavar()} is mutually exclusive with {self._get_other_opts(ctx)}"
                )
            else:
                self.required = False

        return super().handle_parse_result(ctx, opts, args)

    def _get_other_opts(self, ctx):
        for param in ctx.command.params:
            if param.name == self.not_required_if:
                return ' / '.join([f"'{opt}'" for opt in param.opts])


class RequiredIf(click.Option):
    def __init__(self, *args, required_if, **kwargs):
        super().__init__(*args, **kwargs)
        self.required_if = required_if
        assert self.required_if, "'required_if' parameter required"

    def handle_parse_result(self, ctx, opts, args):
        other_present = opts.get(self.required_if) is not None

        if other_present:
            self.required = True

        return super().handle_parse_result(ctx, opts, args)



from pathlib import Path

import click

from kks.util.common import format_file
from kks.util.compat import subprocess


# Используются для поиска скриптов по имени
SCRIPT_EXTENSIONS = ['.py', '.py3', '.sh', '.out', '.cpp']

# Перед вызовом run_script .cpp скрипты будут скомпилированы в временной папке
CPP_EXTENSIONS = ['.cpp']


def run_script(script: Path, args, ignore_exit_code=False, stdin=None, stdout=None, input=None):
    ext = script.suffix

    interpreter = ['python3'] if ext in ['.py', '.py3'] \
        else ['bash'] if ext in ['.sh'] \
        else [] if ext in ['.out', ''] \
        else None

    if interpreter is None:
        click.secho(f'Cant run unrecognized script {format_file(script)}', fg='red', err=True)
        return None

    process = subprocess.run(
        interpreter + [script.absolute()] + args, stdin=stdin, stdout=stdout, input=input
    )

    if process.returncode != 0 and not ignore_exit_code:
        click.echo(
            click.style('Script ', fg='yellow') +
            format_file(script.name) +
            click.style(' exited with code ', fg='yellow') +
            click.style(str(process.returncode), fg='red', bold=True) +
            click.style(' (args: ' + ' '.join(args) + ')', fg='yellow')
        )
        return None

    return process


def needs_compilation(script):
    return script is not None and script.suffix in CPP_EXTENSIONS


def compile_script(workdir, script, options):
    from kks.binary import compile_cpp  # avoid circular import. TODO refactor kks.binary?
    if script.suffix in CPP_EXTENSIONS:
        return compile_cpp(workdir, [script], options)
    else:
        raise Exception(f'Cant compile script with extension {script.suffix}')


def find_script(search_directory, name, default=None, exists=True):
    result = Path(default) if default is not None else None
    if result is None:
        files = search_directory.glob(f'{name}.*')
        for file in files:
            if file.suffix in SCRIPT_EXTENSIONS:
                result = Path(file)
                click.secho(f'Using script {format_file(result)}', err=True)
                break
        if result is None and exists:
            click.secho(f'Can\'t find any script with name "{name}"', fg='red', err=True)
            return None

    if not result.exists() and exists:
        click.secho(f'Script {result} does not exist', fg='red', err=True)
        return None

    return result



import importlib.util
import sys
from functools import wraps


if sys.version_info.major < 3 or sys.version_info.minor < 6:
    raise RuntimeError('Python 3.6+ is required')


__all__ = ['subprocess']


def _module_copy(name):
    # NOTE submodules are shared with the original module
    spec = importlib.util.find_spec(name)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def _subprocess():
    if sys.version_info.minor >= 7:
        import subprocess
        return subprocess
    # don't modify original module
    subprocess = _module_copy('subprocess')
    run = subprocess.run

    @wraps(run)
    def wrapper(*args, **kwargs):
        if kwargs.pop('capture_output', False):
            # backport from 3.7+
            if kwargs.get('stdout') is not None or kwargs.get('stderr') is not None:
                raise ValueError('stdout and stderr arguments may not be used '
                                 'with capture_output.')
            kwargs['stdout'] = subprocess.PIPE
            kwargs['stderr'] = subprocess.PIPE
        return run(*args, **kwargs)

    subprocess.run = wrapper
    return subprocess


subprocess = _subprocess()



import shutil
import tempfile
from pathlib import Path
from sys import exit
from typing import List

import click

from kks.util.common import get_solution_directory, get_clang_style_string, \
                            get_clang_tidy_config, print_diff
from kks.util.compat import subprocess
from kks.util.config import find_target


class SkippedError(Exception):
    pass


@click.command(short_help='Lint solution')
@click.option('--diff/--no-diff', is_flag=True, default=True,
              help='Show lint diff. Always true for dry-run')
@click.option('-n', '--dry-run', is_flag=True, default=False,
              help='Dont actually change any files. Uses temporary directory')
# Not sure if fixes work as intended when applied to multiple files.
# Also some checks (like readability-magic-numbers) just don't have any auto-fixes.
@click.option('-f', '--fix', is_flag=True, default=False,
              help='Apply clang-tidy fixes. Experimental, may break your code. '
                   'Always false for dry-run. If there are compiler errors, fixes are NOT applied.')
@click.option('-T', '-tg', '--target', default='default',
              help='Target name for clang-tidy compiler flags')  # Add --asan/--no-asan?
def lint(diff, dry_run, fix, target):
    """
    Lint solution in current task directory using clang-format and clang-tidy.

    clang-format and clang-tidy must be present in PATH.

    If file ~/.kks/.clang-format exists, it will be passed to clang-format.
    Otherwise, default config (hardcoded in common.py) is used.
    The same applies to clang-tidy.
    """

    directory = get_solution_directory()

    # Also finds C++ generators / reference solutions. Use target.files instead?
    files = (
        list(directory.glob('*.c')) +
        list(directory.glob('*.h')) +
        list(directory.glob('*.cpp'))
    )

    if not files:
        click.secho('No .c, .h, .cpp files found', fg='yellow', err=True)
        return

    all_checks_passed = True
    try:
        if dry_run:
            with tempfile.TemporaryDirectory(prefix='kks-') as work_directory:
                temp_files = [Path(shutil.copy(file, work_directory)) for file in files]
                all_checks_passed &= format_files(temp_files, show_diff=True, diff_error=True)
        else:
            format_ok = format_files(files, show_diff=diff, diff_error=False)
            all_checks_passed &= format_ok
            if format_ok:
                click.secho('Successfully formatted!', fg='green', err=True)
    except SkippedError:
        pass
    try:
        lint_ok = run_clang_tidy(files, target, fix=fix and not dry_run)
        all_checks_passed &= lint_ok
        if lint_ok:
            click.secho('Clang-tidy checks passed!', fg='green', err=True)
    except SkippedError:
        pass
    exit(0 if all_checks_passed else 1)


def _run_binary(args):
    try:
        return subprocess.run(args)
    except FileNotFoundError:
        click.secho(f"'{args[0]}' is not in PATH", fg='yellow', err=True)
        raise SkippedError()


def _get_compiler_flags(source_files: List[Path], target_name: str):
    target = find_target(target_name)
    if target is None:
        click.secho(f'No target {target_name} found', fg='red', err=True)
        return None

    has_c = any(f.name.endswith('.c') for f in source_files)
    has_cpp = any(f.name.endswith('.cpp') for f in source_files)
    if has_c and has_cpp:
        # FIXME clang-tidy is skipped if solution is in C and gen/refsol is in C++
        click.secho('Cannot lint C and C++ together', fg='red', err=True)
        return None

    compiler = target.cpp_compiler if has_cpp else target.compiler
    std = target.cpp_std if has_cpp else target.std
    compiler_args = [compiler, '-std='+std] + target.flags

    if not target.asm64bit and any(f.suffix.lower() == '.s' for f in source_files):
        compiler_args.append('-m32')

    # Add ASAN_ARGS, filenames, libs? clang-tidy seems to work correctly without all these options
    return compiler_args


def format_files(files: List[Path], show_diff: bool, diff_error: bool) -> bool:
    """Runs clang-format on specified files.

    Args:
        files: A list of files to be formatted.
        show_diff: If True, print diff produced by clang-format.
        diff_error: If True, treat non-zero diff as error (return False).

    Returns: True if files are ok (no problems were found), False otherwise.

    Raises:
        SkippedError: clang-format is not in PATH.
    """
    before = {}

    if show_diff or diff_error:
        for file in files:
            with file.open('r') as f:
                before[file] = f.read()

    file_names = [file.as_posix() for file in files]

    files_string = click.style(' '.join(file_names), fg='blue', bold=True)
    click.secho('Formatting files ' + files_string)

    style_string = '--style=' + get_clang_style_string()
    process = _run_binary(['clang-format', '-i', style_string] + file_names)

    if process.returncode != 0:
        click.secho(f'Clang-format exited with code {process.returncode}', fg='red', err=True)
        return False

    if diff_error:
        diff_found = False

    if show_diff or diff_error:
        for file in files:
            with file.open('r') as f:
                after = f.read()
            if diff_error:
                diff_found |= after != before[file]
            if show_diff:
                print_diff(before[file], after, file.as_posix(), file.as_posix())

    if diff_error and diff_found:
        return False
    return True


def run_clang_tidy(files: List[Path], target: str, fix: bool):
    """Runs clang-tidy on specified files.

    Args:
        files: A list of files to be checked.
        target: Name of build target.

    Returns: True if files are ok (no problems were found), False otherwise.

    Raises:
        SkippedError: clang-tidy is not in PATH.
    """
    file_names = [file.as_posix() for file in files]

    files_string = click.style(' '.join(file_names), fg='blue', bold=True)
    click.secho('Linting files ' + files_string)

    compiler_flags = _get_compiler_flags(files, target)
    if compiler_flags is None:
        # internal error
        return True
    clang_tidy_args = ['clang-tidy', get_clang_tidy_config()]
    if fix:
        clang_tidy_args.append('--fix')
    process = _run_binary(clang_tidy_args + file_names + ['--'] + compiler_flags)

    if process.returncode not in [0, 1]:
        # killed / segfault / ???
        click.secho(f'Clang-tidy exited with code {process.returncode}', fg='red', err=True)
        return False

    return process.returncode == 0



from pathlib import Path

import click

from kks.util.click import OptFlagCommand, FlagOption, OptFlagOption, Choice2
from kks.util.common import find_workspace, get_hidden_dir, format_file
from kks.util.config import target_file, global_comment


@click.command(cls=OptFlagCommand)
@click.option('-f', '--force', is_flag=True)
@click.option('-c', '--config', cls=FlagOption, is_flag=True,
              help=f'Create {target_file} in current directory and exit')
@click.option('--config_opt', cls=OptFlagOption, type=Choice2(['update', 'global']),
              help='Create a copy of config for manual updating '
                   'or create config in the root dir of workspace')  # TODO multiline help?
def init(force, config, config_opt):
    """Initialize kks workspace in current directory."""

    path = Path()
    file = path / '.kks-workspace'
    old_workspace = False

    if config or config_opt:
        if config_opt == 'global':
            workspace = find_workspace()
            if workspace is None:
                click.secho(
                    'Current directory is not in a kks workspace.'
                    'To use "global" option, you need to cd into an existing workspace '
                    'or run "kks init"',
                    fg='red'
                )
                return
            path = workspace
            is_global = True
        else:
            is_global = file.exists()
        create_config(path, is_global, config_opt == 'update', force)
        return

    if not force:
        workspace = find_workspace()
        if workspace is not None:
            click.secho(
                f'Found workspace in directory {workspace}\n'
                'If you are sure you want to create workspace here, specify --force',
                fg='yellow'
            )
            return

    if file.exists():
        if file.is_file():
            click.secho('Workspace already exists.', fg='green')
            click.secho('Adding missing files...\n', fg='green')
            old_workspace = True
        elif file.is_dir():
            click.secho('Workspace marker is a directory (.kks-workspace)', fg='yellow')
            return
        else:
            click.secho('Workspace marker is an unknown file type (.kks-workspace)', fg='yellow')
            return
    else:
        file.write_text('This file is used to find kks workspace.\n')

    hidden = get_hidden_dir(path)
    if hidden.exists():
        if hidden.is_dir():
            click.secho(
                f'The directory for hidden contests already exists ({hidden.relative_to(path)})\n',
                fg='yellow'
            )
        else:
            click.secho(
                f'{hidden.relative_to(path)} exists and is not a directory, '
                '"kks (un)hide" may break!\n',
                fg='red'
            )
    else:
        hidden.mkdir()

    action = 'Updated' if old_workspace else 'Initialized'
    click.secho(f'{action} workspace in directory {path.absolute()}', fg='green', bold=True)


def create_config(directory, is_global, update, force):
    from pkg_resources import resource_stream
    file = directory / target_file
    if update:
        file = file.with_suffix(file.suffix + '.default')
    else:
        if file.exists():
            if not file.is_file():
                click.secho(
                    f'"{file}" exists and is not a file, '
                    'you will need to replace it to create custom targets',
                    fg='red'
                )
                return
            elif not force:
                click.secho(
                    f'The file {file} already exists. '
                    'Use --force to overwrite it or --config=update to create a copy',
                    fg='yellow'
                )
                return

    data = resource_stream('kks', f'data/{target_file}').read().decode()
    if is_global:
        data = global_comment + data
    file.write_text(data)

    if update:
        click.secho('New default targets are written to ', fg='green', nl=False)
        click.secho(format_file(file), nl=False)
        click.secho(', you can merge it with ', fg='green', nl=False)
        click.secho(format_file(file.with_name(target_file)), nl=False)
        click.secho(' manually', fg='green')
    else:
        click.secho('Default targets are written to ', fg='green', nl=False)
        click.secho(format_file(file))
        click.secho('The config file is not updated automatically.', bold=True)
